@@ -9,20 +9,21 @@ __all__ = [
 
 import errno
 
+from collections.abc import Callable, Mapping
 from functools import cached_property, partial
 from http.client import HTTPResponse
 from io import (
     BufferedReader, RawIOBase, TextIOWrapper, UnsupportedOperation, DEFAULT_BUFFER_SIZE, 
 )
 from os import fstat, stat, PathLike
-from typing import Any, BinaryIO, Callable, Mapping, Optional, Protocol, TypeVar
+from typing import Any, BinaryIO, IO, Optional, Protocol, Self, TypeVar
 from types import MappingProxyType
 from warnings import warn
 
 from requests import Session
 
 from .property import funcproperty
-from .response import get_filename, get_length, is_chunked, is_range_request
+from .response import get_filename, get_range, get_total_length, is_chunked, is_range_request
 from .urlopen import urlopen
 
 
@@ -112,7 +113,7 @@ def get_filesize(file, /) -> int:
 
 
 class HTTPFileReader(RawIOBase):
-    url: str
+    url: str | Callable[[], str]
     response: Any
     length: int
     chunked: bool
@@ -124,19 +125,26 @@ class HTTPFileReader(RawIOBase):
 
     def __init__(
         self, 
-        url: str, 
+        url: str | Callable[[], str], 
         /, 
         headers: Mapping = {}, 
+        start: int = 0, 
         urlopen: Callable[..., HTTPResponse] = urlopen, 
     ):
         headers = {**headers, "Accept-Encoding": "identity"}
-        response = urlopen(url, headers=headers)
+        if start >= 0:
+            headers["Range"] = f"bytes={start}-"
+        else:
+            headers["Range"] = f"bytes={start}"
+        response = urlopen(url() if callable(url) else url, headers=headers)
+        rng = get_range(response)
+        start = rng[0] if rng else 0
         self.__dict__.update(
             url = url, 
             response = response, 
-            length = get_length(response) or 0, 
+            length = get_total_length(response) or 0, 
             chunked = is_chunked(response), 
-            start = 0, 
+            start = start, 
             closed = False, 
             urlopen = urlopen, 
             headers = MappingProxyType(headers), 
@@ -187,7 +195,7 @@ class HTTPFileReader(RawIOBase):
         self.__dict__["closed"] = True
 
     @funcproperty
-    def file(self, /):
+    def file(self, /) -> BinaryIO:
         return self.response
 
     @funcproperty
@@ -215,7 +223,10 @@ class HTTPFileReader(RawIOBase):
             return b""
         if self.file.closed:
             self.reconnect()
-        data = self.file.read(size)
+        if size is None or size < 0:
+            data = self.file.read()
+        else:
+            data = self.file.read(size)
         if data:
             self._add_start(len(data))
         return data
@@ -226,6 +237,8 @@ class HTTPFileReader(RawIOBase):
     def readinto(self, buffer, /) -> int:
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if not self.chunked and self.tell() >= self.length:
+            return 0
         if self.file.closed:
             self.reconnect()
         size = self.file.readinto(buffer)
@@ -238,11 +251,12 @@ class HTTPFileReader(RawIOBase):
             raise ValueError("I/O operation on closed file.")
         if size == 0 or not self.chunked and self.tell() >= self.length:
             return b""
-        elif size is None:
-            size = -1
         if self.file.closed:
             self.reconnect()
-        data = self.file.readline(size)
+        if size is None or size < 0:
+            data = self.file.readline()
+        else:
+            data = self.file.readline(size)
         if data:
             self._add_start(len(data))
         return data
@@ -250,6 +264,8 @@ class HTTPFileReader(RawIOBase):
     def readlines(self, hint: int = -1, /) -> list[bytes]:
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if not self.chunked and self.tell() >= self.length:
+            return []
         if self.file.closed:
             self.reconnect()
         ls = self.file.readlines(hint)
@@ -268,9 +284,16 @@ class HTTPFileReader(RawIOBase):
             start = self.length + start
             if start < 0:
                 start = 0
+        if start >= self.length:
+            self.__dict__.update(start=start)
+            return start
         self.response.close()
+        url = self.url
         self.__dict__.update(
-            response=self.urlopen(self.url, headers={**self.headers, "Range": f"bytes={start}-"}), 
+            response=self.urlopen(
+                url() if callable(url) else url, 
+                headers={**self.headers, "Range": f"bytes={start}-"}
+            ), 
             start=start, 
         )
         return start
@@ -321,7 +344,7 @@ class HTTPFileReader(RawIOBase):
         encoding: Optional[str] = None, 
         errors: Optional[str] = None, 
         newline: Optional[str] = None, 
-    ):
+    ) -> Self | IO:
         if buffering is None:
             if text_mode:
                 buffering = DEFAULT_BUFFER_SIZE
@@ -361,32 +384,28 @@ class RequestsFileReader(HTTPFileReader):
 
     def __init__(
         self, 
-        url: str, 
+        url: str | Callable[[], str], 
         /, 
         headers: Mapping = {}, 
+        start: int = 0, 
         urlopen: Callable = Session().get, 
     ):
-        headers = {**headers, "Accept-Encoding": "identity"}
-        response = urlopen(url, headers=headers, stream=True)
-        self.__dict__.update(
-            url = url, 
-            response = response, 
-            length = get_length(response) or 0, 
-            chunked = is_chunked(response), 
-            start = 0, 
-            closed = False, 
-            urlopen = partial(urlopen, stream=True), 
-            headers = MappingProxyType(headers), 
-            _seekable = is_range_request(response), 
-        )
+        def urlopen_wrapper(url: str, headers: Mapping = headers):
+            resp = urlopen(url, headers=headers, stream=True)
+            resp.raise_for_status()
+            return resp
+        super().__init__(url, headers=headers, start=start, urlopen=urlopen_wrapper)
 
     def _add_start(self, delta: int, /):
         pass
 
     @funcproperty
-    def file(self, /):
+    def file(self, /) -> BinaryIO:
         return self.response.raw
 
     def tell(self, /) -> int:
-        return self.start + self.file.tell()
+        start = self.start
+        if start >= self.length:
+            return start
+        return start + self.file.tell()
 
